# coding: utf-8

import requests
from flask_srd.consul_dns import ConsulResolver
import random
from flask import current_app as app
from urllib.parse import urljoin
from flask_srd.decorators import with_retry_connections


class ConsulService(object):
    def __init__(self, service_name, tag=None, nameservers=None, port=None, domain='service.consul'):
        """
        :param service_name: str
        :param nameservers: consul dns server list
        :param port: dns port
        :param domain: str
        """
        if not isinstance(nameservers, list):
            raise ValueError('NameServer must be a list')

        self.service_name = service_name
        self.tag = tag
        self.domain = domain
        self.resolver = ConsulResolver(nameservers=nameservers, port=port)
        self.session = requests.Session()

    def _resolve(self):
        """
        Query the consul DNS server for the service IP and port
        """
        services = self.resolver.query(service_name=self.service_name, tag=self.tag, domain=self.domain)
        app.logger.info(f'Available Services {self.service_name}: {services}')
        choose_service = random.choice(services)
        app.logger.info(f'Choose Server: {choose_service.host}:{choose_service.port}')
        return f'http://{choose_service.host}:{choose_service.port}'

    @property
    def base_url(self):
        """
        get the next endpoint from self.endpoints
        """
        return self._resolve()

    @with_retry_connections()
    def request(self, method, endpoint, **kwargs):
        """
        Proxy to requests.request
        :param method: str formatted http method
        :param endpoint: service endpoint
        :param kwargs: kwargs passed directly to requests.request
        :return:
        """
        kwargs.setdefault('timeout', (1, 30))
        return self.session.request(
            method,
            urljoin(self.base_url, endpoint),
            **kwargs
        )

    def get(self, endpoint, **kwargs):
        return self.request('GET', endpoint, **kwargs)

    def post(self, endpoint, **kwargs):
        return self.request('POST', endpoint, **kwargs)

    def delete(self, endpoint, **kwargs):
        return self.request('DELETE', endpoint, **kwargs)

    def put(self, endpoint, **kwargs):
        return self.request('PUT', endpoint, **kwargs)

    def options(self, endpoint, **kwargs):
        return self.request('OPTIONS', endpoint, **kwargs)

    def head(self, endpoint, **kwargs):
        return self.request('HEAD', endpoint, **kwargs)



import os
import json
import consulate

from flask_srd.decorators import with_retry_connections


class Consul(object):
    """
    The Consul flask.ext object is responsible for connecting and querying
    consul (using gmr/consulate as the underlying client library).
    """

    def __init__(self, app=None, **kwargs):
        """
        Initialize the flask extension

        :param app: flask.Flask application instance
        :param kwargs:
            consul_host: host to connect to, falling back to environmental
                        variable $CONSUL_HOST, then 'localhost'
            consul_port: port, falling back to $CONSUL_PORT, then 8500
            healthcheck: healthcheck that will be registered
            max_tries: integer number of attempts to make to connect to
                        consul_host. Useful if the host is an alias for
                        the consul cluster
        :return: None
        """
        self.kwargs = kwargs if kwargs else {}
        self.app = None

        self.host = self.kwargs.get('consul_host') or os.environ.get('CONSUL_HOST', 'localhost')
        self.port = self.kwargs.get('consul_port') or os.environ.get('CONSUL_PORT', 8500)
        self.max_tries = self.kwargs.get('max_tries', 3)

        self.session = None

        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        self.app = app

        if not hasattr(app, 'extensions'):
            app.extensions = {}
        if 'consul' in app.extensions:
            raise RuntimeError('Flask application already initialized')
        app.extensions['consul'] = self

        self.session = self._create_session(
            test_connection=self.kwargs.get('test_connection', False),
        )

    @with_retry_connections()
    def _create_session(self, test_connection=False):
        """
        Create a consulate.session object, and query for its leader to ensure
        that the connection is made.

        :param test_connection: call .leader() to ensure that the connection
            is valid
        :type test_connection: bool
        :return consulate.Session instance
        """
        session = consulate.Session(host=self.host, port=self.port)
        if test_connection:
            session.status.leader()
        return session

    @with_retry_connections()
    def apply_remote_config(self, namespace=None):
        """
        Applies all config values defined in consul's kv store to self.app.

        There is no guarantee that these values will not be overwritten later
        elsewhere.

        :param namespace: kv namespace/directory
        :return: None
        """
        self.app.logger.info('Apply remote config')
        try:
            kv = self.session.kv.find(namespace)
            for k, v in json.loads(kv[namespace]).items():
                self.app.config[k] = v
                msg = "Set {k}={v} from consul kv '{ns}'".format(
                    k=k,
                    v=v,
                    ns=namespace,
                )
                self.app.logger.info(msg)
        except KeyError:
            self.app.logger.warning(f'{namespace} namespace is not found')
        except (TypeError, ValueError):
            self.app.logger.warning('KV is not json-serializable')

    @with_retry_connections()
    def register_service(self, **kwargs):
        """
        register this service with consul
        kwargs passed to Consul.agent.service.register
        """
        kwargs.setdefault('name', self.app.name)
        self.session.agent.service.register(**kwargs)



"""
flask-consulate
-------------
flask extension that provides an interface to consul via a flask.app
"""

from setuptools import setup, find_packages

INSTALL_REQUIRES = [
    'consulate',
    'Flask',
    'requests>=2.7.0',
    'dnspython',
]

setup(
    name='flask-srd',
    version='0.0.1',
    url='http://github.com/megachweng/flask-srd',
    license='MIT',
    author='megachweng',
    author_email='megachweng@gmail.com',

    description='flask extension that provides service registry and discovery functionality',
    long_description=__doc__,

    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=INSTALL_REQUIRES,

    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



import collections
import logging
from dns import rdatatype
from dns.resolver import Resolver
from dns.resolver import NoNameservers, NoAnswer

LOGGER = logging.getLogger(__name__)

SRV = collections.namedtuple(
    'SRV', ['host', 'port', 'priority', 'weight', 'hostname'])


class ConsulResolver:
    def __init__(self, nameservers: list = None, port: int = 8600):
        self.resolver = Resolver()
        self.resolver.nameservers = nameservers
        self.resolver.port = port

    def query(self, service_name: str, tag: str = None, domain: str = 'service.consul'):
        label = [q for q in [tag, service_name, domain] if q]
        try:
            answer = self.resolver.query('.'.join(label), 'SRV')
            return self._build_result_set(answer)
        except (NoNameservers, NoAnswer) as e:
            LOGGER.error(f'DNS Query Failed :{e.msg}')

    @staticmethod
    def _build_result_set(answer):
        """Return a list of SRV instances for a DNS answer.
        :rtype: list of SRV
        """

        def _build_resource_to_address_map(answer):
            mapping = collections.defaultdict(list)
            for resource in answer.response.additional:
                target = resource.name.to_text()
                mapping[target].extend(record.address for record in resource.items if record.rdtype == rdatatype.A)
            return mapping

        resource_map = _build_resource_to_address_map(answer)
        result_set = []
        for resource in answer:
            target = resource.target.to_text()
            if target in resource_map:
                result_set.extend(
                    SRV(address, resource.port, resource.priority, resource.weight,
                        target.strip('.')) for address in resource_map[target])
            else:
                result_set.append(
                    SRV(target.rstrip('.'), resource.port, resource.priority,
                        resource.weight, target.strip('.')))
        return result_set


if __name__ == '__main__':
    r = ConsulResolver(nameservers=['10.1.1.7'])
    print(r.query(tag='python', service_name='a', domain='service.consul'))
